package java.lang.invoke;

import java.lang.reflect.Field;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.lang.reflect.Member;

final class InfoFromMemberName implements MethodHandleInfo
{
    private final MemberName member;
    private final int referenceKind;
    
    InfoFromMemberName(MethodHandles.Lookup lookup, MemberName member, byte referenceKind) {
        assert !(!member.isMethodHandleInvoke());
        assert !(!member.referenceKindIsConsistentWith((int)referenceKind));
        this.member = member;
        this.referenceKind = referenceKind;
        return;
    }
    
    @Override
    public Class<?> getDeclaringClass() {
        return this.member.getDeclaringClass();
    }
    
    @Override
    public String getName() {
        return this.member.getName();
    }
    
    @Override
    public MethodType getMethodType() {
        return this.member.getMethodOrFieldType();
    }
    
    @Override
    public int getModifiers() {
        return this.member.getModifiers();
    }
    
    @Override
    public int getReferenceKind() {
        return this.referenceKind;
    }
    
    @Override
    public String toString() {
        return MethodHandleInfo.toString(this.getReferenceKind(), this.getDeclaringClass(), this.getName(), this.getMethodType());
    }
    
    @Override
    public <T extends Member> T reflectAs(Class<T> clazz, MethodHandles.Lookup lookup) {
        if (this.member.isMethodHandleInvoke() && !this.member.isVarargs()) {
            throw new IllegalArgumentException("cannot reflect signature polymorphic method");
        }
        Member obj = (Member)AccessController.doPrivileged(new PrivilegedAction<Member>() {
            @Override
            public Member run() {
                try {
                    return InfoFromMemberName.access$000(InfoFromMemberName.this);
                }
                catch (ReflectiveOperationException cause) {
                    throw new IllegalArgumentException(cause);
                }
            }
        });
        try {
            Class<Object> declaringClass = this.getDeclaringClass();
            int n = (byte)this.getReferenceKind();
            lookup.checkAccess((byte)n, declaringClass, convertToMemberName((byte)n, obj));
        }
        catch (IllegalAccessException cause) {
            throw new IllegalArgumentException(cause);
        }
        return (T)clazz.cast(obj);
    }
    
    private Member reflectUnchecked() throws ReflectiveOperationException {
        int i = (byte)this.getReferenceKind();
        Class<Object> declaringClass = this.getDeclaringClass();
        int public1 = Modifier.isPublic(this.getModifiers());
        if (MethodHandleNatives.refKindIsMethod((byte)i)) {
            if (!(public1 == 0)) {
                return declaringClass.getMethod(this.getName(), this.getMethodType().parameterArray());
            }
            return declaringClass.getDeclaredMethod(this.getName(), this.getMethodType().parameterArray());
        }
        else if (MethodHandleNatives.refKindIsConstructor((byte)i)) {
            if (!(public1 == 0)) {
                return declaringClass.getConstructor(this.getMethodType().parameterArray());
            }
            return declaringClass.getDeclaredConstructor(this.getMethodType().parameterArray());
        }
        else if (MethodHandleNatives.refKindIsField((byte)i)) {
            if (!(public1 == 0)) {
                return declaringClass.getField(this.getName());
            }
            return declaringClass.getDeclaredField(this.getName());
        }
        else {
            throw new IllegalArgumentException("referenceKind=" + i);
        }
    }
    
    private static MemberName convertToMemberName(byte b, Member member) throws IllegalAccessException {
        if ((member instanceof Method)) {
            return new MemberName((Method)member, b == 7);
        }
        if ((member instanceof Constructor)) {
            return new MemberName((Constructor)member);
        }
        if ((member instanceof Field)) {
            return new MemberName((Field)member, b == 3 || b == 4);
        }
        throw new InternalError(member.getClass().getName());
    }
    
    static Member access$000(InfoFromMemberName infoFromMemberName) throws ReflectiveOperationException {
        return infoFromMemberName.reflectUnchecked();
    }
}
